"""### To import textual_pyfiglet:
from textual_pyfiglet import FigletWidget

### You can also import the original PyFiglet:
from textual_pyfiglet.pyfiglet import Figlet         <-- Class version
from textual_pyfiglet.pyfiglet import figlet_format  <-- Function version

### To install the extended fonts collection:   
pip install textual-pyfiglet[fonts]

You can also download FIG fonts from the internet and just drop them in the fonts folder.   
See the readme for more information.
 
You can access the original PyFiglet CLI with the following command:   
python -m textual_pyfiglet.pyfiglet
"""
from textual_pyfiglet.figletwidget import FigletWidget

from textual_pyfiglet.config import check_for_extended_fonts

# If recently installed, this copies the fonts to the user directory.
are_extended_fonts_installed: bool = check_for_extended_fonts()





"""Module for the FigletWidget class.

Import FigletWidget into your project to use it."""

from __future__ import annotations
import os
from platformdirs import user_data_dir

from textual.message import Message
from textual.widgets import Static
from textual.reactive import reactive

from textual_pyfiglet.pyfiglet import Figlet, fonts
from textual_pyfiglet.config import check_for_extended_fonts

are_extended_fonts_installed: bool = check_for_extended_fonts()


class _InnerFiglet(Static):
    """The Inner Figlet contains the actual PyFiglet object.
    This nested layout is necessary for things to work properly."""

    figlet_input:  reactive[str] = reactive('', always_update=True)
    figlet_output: reactive[str] = reactive('', layout=True)
    font:          reactive[str] = reactive('standard')

    # Note: Indeed its a bit odd that justify is not also a reactive... I had to go into
    # the pyfiglet source code to create a setter method for justify. (It previously
    # only had a getter method). It seems to work identically without being reactive.
    # I don't know why.

    def __init__(self, *args, font, justify, **kwargs) -> None:
        """Private class for the FigletWidget. Same as Static except for `font` and `justify` args.   
        Don't use this class. Use FigletWidget instead. """
        
        super().__init__(*args, **kwargs)
        self.figlet = Figlet(font=font, justify=justify)
        self.font = font

    def watch_font(self, value: str) -> None:
        self.figlet.setFont(font=value)
        self.watch_figlet_input(self.figlet_input)

    def watch_figlet_input(self, value: str) -> None:

        if not self.parent or self.parent.size.width == 0:
            return
        self.figlet.width = self.parent.size.width

        if value == '':
            self.figlet_output = ''
        else:
            self.figlet_output = self.figlet.renderText(value)

    def render(self) -> str:
        return self.figlet_output


class FigletWidget(Static):
    """Adds PyFiglet ability to the Static widget.
    
    See init docstring for more info."""

    DEFAULT_CSS = """
    FigletWidget {
        width: auto;
        height: auto;
        padding: 0;
    }
    """

    are_extended_fonts_installed = are_extended_fonts_installed
    """This is a convenience variable, so the class can easily check 
    if the extended fonts pack has been installed."""

    base_fonts = [
        'calvin_s',
        'chunky',
        'cybermedium',
        'small_slant',
        'small',
        'smblock',
        'smbraille',
        'standard',
        'stick_letters',
        'tmplr'
    ]

    class Updated(Message):
        """This is here to provide a message to the app that the widget has been updated.
        You might need this to trigger something else in your app resizing, adjusting, etc.
        The size of FIG fonts can vary greatly, so this might help you adjust other widgets."""

        def __init__(self, widget: FigletWidget) -> None:
            super().__init__()
            self.widget = widget
            '''The FigletWidget that was updated.'''

        @property
        def control(self) -> FigletWidget:
            return self.widget


    def __init__(self, *args, font: str = "calvin_s", justify: str = "center", **kwargs) -> None:
        """A custom widget for turning text into ASCII art using PyFiglet.
        This args section is copied from the Static widget. It's the same except for the font argument.

        This class is designed to be an easy drop in replacement for the Static widget.
        The only new argument is 'font', which has a default set to one of the smallest fonts.
        You can replace any Static widget with this and it should work (aside from the size).

        The widget will try to adjust its render area to fit inside of its parent container.
        The easiest way to use this widget is to place it inside of a container.
        Resize the parent container, and then call the `update()` method.

        Args:
            renderable: A Rich renderable, or string containing console markup.
            font (PyFiglet): Font to use for the ASCII art. Default is "calvin_s".
            justify (PyFiglet): Justification for the text. Default is "center".
            expand: Expand content if required to fill container.
            shrink: Shrink content if required to fill container.
            markup: True if markup should be parsed and rendered.
            name: Name of widget.
            id: ID of Widget.
            classes: Space separated list of class names.
            disabled: Whether the static is disabled or not.

        Included fonts:
        - calvin_s
        - chunky
        - cybermedium
        - small_slant
        - small
        - smblock
        - smbraille
        - standard
        - stick_letters
        - tmplr

        Remember you can download more fonts. To download the extended fonts pack:
        pip install textual-pyfiglet[fonts]
        """
        super().__init__(*args, **kwargs)
        self.stored_text = str(self.renderable)
        self.renderable = ''
        self.font = font
        self.justify = justify

        # NOTE: Figlet also has a "direction" argument
        # TODO Add Direction argument

    def compose(self):
        self._inner_figlet = _InnerFiglet(
            renderable = self.stored_text,
            id='inner_figlet',
            font=self.font,
            justify=self.justify
        )
        yield self._inner_figlet

    def on_mount(self):
        self.update(self.stored_text)

    def on_resize(self):
        self.update()

    def update(self, new_text: str | None = None) -> None:
        '''Update the PyFiglet area with the new text.    
        Note that this over-rides the standard update method in the Static widget.   
        Unlike the Static widget, this method does not take a Rich renderable.   
        It can only take a text string. Figlet needs a normal string to work properly.

        Args:
            new_text: The text to update the PyFiglet widget with. Default is None.'''
        
        if new_text is not None:
            self.stored_text = new_text

        # self._inner_figlet.update(self.stored_text)
        self._inner_figlet.figlet_input = self.stored_text
        self.post_message(self.Updated(self))

    def set_font(self, font: str) -> None:
        """Set the font for the PyFiglet widget.   
        The widget will update with the new font automatically.
        
        Pass in the name of the font as a string:
        ie 'calvin_s', 'small', etc.
        
        Args:
            font: The name of the font to set."""
        
        # self._inner_figlet.figlet.setFont(font=font)
        self._inner_figlet.font = font
        self.update()

    def set_justify(self, justify: str) -> None:
        """Set the justification for the PyFiglet widget.   
        The widget will update with the new justification automatically.
        
        Pass in the justification as a string:   
        options are: 'left', 'center', 'right', 'auto'
        
        Args:
            justify: The justification to set."""
        
        self._inner_figlet.figlet.justify = justify
        self.update()

    def get_fonts_list(self, get_all: bool = True) -> list:
        """Scans the fonts folder.   
        Returns a list of all font filenames (without extensions).
        
        Args:
            get_all: If True, returns all fonts. If False, returns only the base fonts."""

        if not get_all:
            return self.base_fonts

        # first get the path of the fonts package:
        base_fonts_folder = os.path.dirname(fonts.__file__)
        base_fonts_folder_contents = os.listdir(base_fonts_folder)    # list of all files in the fonts folder
        fonts_list = []

        for filename in base_fonts_folder_contents:
            if filename.endswith('.flf') or filename.endswith('.tlf'):
                fonts_list.append(os.path.splitext(filename)[0])

        user_fonts_folder = user_data_dir('pyfiglet', appauthor=False)
        user_fonts_folder_contents = os.listdir(user_fonts_folder)

        for filename in user_fonts_folder_contents:
            if filename.endswith('.flf') or filename.endswith('.tlf'):
                fonts_list.append(os.path.splitext(filename)[0])

        return fonts_list
    
    def copy_figlet_to_clipboard(self) -> None:
        """Copy the PyFiglet text to the clipboard."""
        self.app.copy_to_clipboard(str(self._inner_figlet.renderable))

    def return_figlet_as_string(self) -> str:
        """Return the PyFiglet text as a string."""
        return str(self._inner_figlet.renderable)
    



"""Contains the demo app.
This module contains the demo application for PyFiglet

Run this module directly to start the PyFiglet demo application.    
Either command works:   
- textual-pyfiglet
- python -m textual-pyfiglet

Note the fonts list is dynamically generated from the fonts folder.
If you also installed the extended fonts pack, You can toggle the switch
in the demo, and it will scan the fonts folder for all fonts.
(You can add your own, as well, or download them individually.)

Also note that the first time it runs and detects the extended fonts pack,
it will copy the whole pack into the fonts folder (inside pyfiglet).
This might take about 1-2 seconds. It will only do this once.
"""
# Python imports
from typing import cast

# Textual imports
from textual.app import App, on
from textual.containers import Horizontal, Vertical, Container, VerticalScroll
from textual.widgets import (
    Header,
    Footer,
    Button,
    Static,
    Input,
    TextArea,
    Select,
    Switch,
    Label,
)

# textual-pyfiglet imports
from textual_pyfiglet.figletwidget import FigletWidget
from textual_pyfiglet.pyfiglet import figlet_format
from textual_pyfiglet import are_extended_fonts_installed 

from rich import traceback
traceback.install()             # this has to be last or Ruff / linter complains


class TextualPyFigletDemo(App):

    BINDINGS = [
        ("s", "focus_select", "Focus the font select"),
        ("t", "focus_text", "Focus the text input"),
        ("a", "toggle_fonts", "Toggle all fonts"),
    ]

    CSS_PATH = "styles.tcss"
    current_font = 'standard'       # starting font for the demo
    justifications = [
        ('Left', 'left'),
        ('Center', 'center'),
        ('Right', 'right'),
    ]

    # timers, used for notifications
    timer1 = None
    timer2 = None

    def compose(self):

        self.log("PyFiglet Demo started.")
        self.log(figlet_format("Textual-\nPyFiglet", font="standard"))
        self.log(f"Extended fonts installed: {are_extended_fonts_installed}")

        yield Header("PyFiglet Demo")

        with Horizontal():
            with Vertical(id="sidebar"):
                yield Label("Set container width:", classes="sidebar_label")
                yield Input(id="width_input", classes="sidebar_input")
                yield Label("Set container height:", classes="sidebar_label")
                yield Input(id="height_input", classes="sidebar_input")
                yield Label("\n(Leave blank\n for auto)\n", classes="sidebar_label")
                yield Label("Justify:", classes="sidebar_label") 
                yield Select(self.justifications, id="justify_select", value='center', allow_blank=False)
                yield Button("Set", id="set_button", classes="sidebar_button")
                yield Button("Copy text\nto clipboard", id="copy_button", classes="sidebar_button")

            with VerticalScroll(id="main_window"):
                yield FigletWidget("Starter Text", id="figlet_widget")

        with Container(id="bottom_bar"):
            with Horizontal():
                yield Static(id="notification_box1")
                yield Static(id="notification_box2")
            with Horizontal():
                yield Select([], id="font_select", allow_blank=True)
                with Horizontal(id="container_of_switch"):
                    yield Switch(False, id="switch")
                    yield Label("Show all \nfonts")
                yield TextArea(id="text_input")

        yield Footer()

    def on_mount(self):

        self.figlet_widget = cast(FigletWidget, self.query_one("#figlet_widget")) 
        self.font_select   = cast(Select, self.query_one("#font_select"))            
        self.text_input    = cast(TextArea, self.query_one("#text_input"))    # chad type hinting convenience vars
        self.font_switch   = cast(Switch, self.query_one("#switch"))
        self.notification1 = cast(Label, self.query_one("#notification_box1"))
        self.notification2 = cast(Label, self.query_one("#notification_box2"))
        self.width_input   = cast(Input, self.query_one("#width_input"))
        self.height_input  = cast(Input, self.query_one("#height_input"))

        self.figlet_widget._inner_figlet.tooltip = "Inner Figlet Widget"
        self.figlet_widget.tooltip = "Outer Figlet Widget"

        self.fonts_list = self.figlet_widget.get_fonts_list(get_all=True)
        self.base_fonts = self.figlet_widget.get_fonts_list(get_all=False)
        self.fonts_list.sort()
        self.font_options = [(font, font) for font in self.base_fonts]
        self.font_select.set_options(self.font_options)
        self.font_select.value = self.current_font          # NOTE: This is setting the font.

        self.text_input.focus()
        end = self.text_input.get_cursor_line_end_location()
        self.text_input.move_cursor(end)

    # Because the select box has a default value, this will run on startup, and then set
    # the font to the default selection. You can also set a font in the constructor of the FigletWidget,
    # But for the purposes of the demo I'm not doing that here because this would override the constructor setting.

    @on(Select.Changed, selector="#font_select")           
    def font_changed(self, event: Select.Changed) -> None:
        if event.value == Select.BLANK:
            return
        self.figlet_widget.set_font(event.value)
        self.current_font = event.value
        self.log(f"Current font set to: {self.current_font}")

    # why this no work?
    # def on_key(self, event):
    #     if event.key == "up":
    #         self.font_select.scroll_up()
    #     elif event.key == "down":
    #         self.font_select.scroll_down()

    @on(Select.Changed, selector="#justify_select")
    def justify_changed(self, event: Select.Changed) -> None:
        self.figlet_widget.set_justify(event.value)
        self.log(f"Justify set to: {event.value}")

    @on(Switch.Changed)
    def toggle_fonts(self, event: Switch.Changed) -> None:
        if event.value:                                         # turn on extended
            self.font_options = [(font, font) for font in self.fonts_list]
        else:
            self.font_options = [(font, font) for font in self.base_fonts]
        self.font_select.set_options(self.font_options)

        if not event.value:                                     # if turning switch off:                         
            if self.current_font not in self.base_fonts:        # if the current font is not in the base fonts list,
                self.current_font = 'standard'                  # set back to standard.
                self.figlet_widget.set_font('standard')         # Keeps font the same when toggling switch, if it can.
        self.font_select.value = self.current_font

        if event.value:
            if are_extended_fonts_installed:
                self.show_notification1("Scanning folder... Extended fonts detected.")
            else:
                self.show_notification1("Scanning folder...")


    @on(Button.Pressed, selector="#set_button")
    def set_container_size(self):
        width = self.width_input.value
        height = self.height_input.value
        self.log(f"Setting container size to: ({width} x {height})")
        if width:
            self.figlet_widget.styles.width = int(width)
        else:
            self.figlet_widget.set_styles('width: 1fr;')
        if height:
            self.figlet_widget.styles.height = int(height)
        else:
            self.figlet_widget.set_styles('height: auto;')

    @on(Button.Pressed, selector="#copy_button")
    def copy_text(self):
        self.figlet_widget.copy_figlet_to_clipboard()

    @on(FigletWidget.Updated)
    def figlet_updated(self, event: FigletWidget.Updated):
        outer_width, outer_height = event.widget.size
        inner_width, inner_height = event.widget._inner_figlet.size
        self.show_notification2(f"Outer: {outer_width}W x {outer_height}H | Inner: {inner_width}W x {inner_height}H")

    @on(TextArea.Changed)
    async def text_updated(self):
        text = self.text_input.text
        self.figlet_widget.update(text)

        # This just scrolls the text area to the end when the text changes:
        scroll_area = self.query_one("#main_window")
        if scroll_area.scrollbars_enabled == (True, False):
            scroll_area.action_scroll_end()

    def show_notification1(self, message: str):
        self.notification1.update(message)
        if self.timer1:
            self.timer1.stop()
        self.timer1 = self.set_timer(3, self.clear_notification1)

    def show_notification2(self, message: str):
        self.notification2.update(message)
        # if self.timer2:
        #     self.timer2.stop()
        # self.timer2 = self.set_timer(3, self.clear_notification2)

    def clear_notification1(self):
        self.notification1.update('')
        self.timer = None

    def clear_notification2(self):
        self.notification2.update('')
        self.timer = None

    def action_focus_select(self):
        self.font_select.focus()

    def action_focus_text(self):
        self.text_input.focus()

    def action_toggle_fonts(self):
        self.font_switch.value = not self.font_switch.value

# This is for the entry point script. You can run the demo with:
#$ textual-pyfiglet
def main():
    app = TextualPyFigletDemo()
    app.run()
 
# This is another way to run the demo
#$ python -m textual-pyfiglet 
if __name__ == "__main__":
    app = TextualPyFigletDemo()
    app.run()


"""This function exists to check if the extended fonts pack has been downloaded,
and to copy it to the user data directory if it hasn't been already.

This uses platformdirs to get the user data directory, which is different for each OS."""


import os
import shutil
import importlib
import importlib.resources
from platformdirs import user_data_dir


def check_for_extended_fonts() -> str:
    """This function will check if fonts have been installed.   
    This check will run every time the package is imported, so you can install the fonts at any time.
    
    Returns one of three conditions:
    1) installed - the fonts are already installed.
    2) not_installed - did not detect fonts pack.
    3) just_installed - the fonts were just installed right now. """

    package_name = "textual_pyfiglet_fonts"

    user_fonts_folder = user_data_dir('pyfiglet', appauthor=False)     # get user data dir

    if os.path.exists(os.path.join(user_fonts_folder, '.fonts_installed')):   #look for flag file
        return 'installed'

    # check if the fonts package is installed
    if importlib.util.find_spec(package_name) is not None:
        fonts_module = importlib.import_module(package_name)
    else:
        return 'not_installed'
    
    # if we're here, the fonts package must be recently installed.
    
    fonts_module_path = os.path.dirname(fonts_module.__file__)   # get the path to the fonts package
    fonts_pack_list: list = os.listdir(fonts_module_path)       # get list of files in fonts package

    # copy all files to the user fonts folder
    for font_file in fonts_pack_list:
        if not font_file.endswith('.py') and not font_file.endswith('__pycache__'):
            font_path = os.path.join(fonts_module_path, font_file)
            new_font_path = os.path.join(user_fonts_folder, font_file)
            try:
                shutil.copyfile(font_path, new_font_path)
            except Exception as e:
                print(f"Error copying font: {font_file} - {e}")

    return 'just_installed'